/* file: SparkKmeansCSR.java */
//==============================================================
//
// SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
// http://software.intel.com/en-us/articles/intel-sample-source-code-license-agreement/
//
// Copyright 2016 Intel Corporation
//
// THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT
// NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
// PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL PROPERTY RIGHTS.
//
// =============================================================

/*
//  Content:
//      Java sample of K-Means clustering in the distributed processing mode
////////////////////////////////////////////////////////////////////////////////
*/

package DAAL;

import java.util.ArrayList;
import java.util.List;
import java.util.Arrays;
import java.util.Map;
import java.util.HashMap;

import org.apache.spark.api.java.*;
import org.apache.spark.api.java.function.*;
import org.apache.spark.SparkConf;

import scala.Tuple2;
import com.intel.daal.algorithms.kmeans.*;
import com.intel.daal.algorithms.kmeans.init.*;
import com.intel.daal.data_management.data.*;
import com.intel.daal.services.*;

import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;

public class SparkKmeansCSR {
    /* Class containing the algorithm results */
    static class KmeansResult {
        public HomogenNumericTable centroids;
    }
    private static final long nBlocks   = 1;
    private static final long nClusters = 20;
    private static final int nIterations = 5;
    private static final int nVectorsInBlock = 8000;

    public static KmeansResult runKmeans(DaalContext context, JavaPairRDD<Integer, CSRNumericTable> dataRDD) {

        JavaRDD<InitPartialResult> partsRDD = computeInitLocal(context, dataRDD);
        InitPartialResult initPartRes = reduceInitPartialResults(context, partsRDD);
        HomogenNumericTable centroids = computeInitMaster(context, initPartRes);

        PartialResult partRes;
        JavaRDD<PartialResult> partsRDDcompute;
        for(int it = 0; it < nIterations; it++) {
            centroids.pack();
            partsRDDcompute = computeLocal(context, dataRDD, centroids);
            partRes = reducePartialResults(context, partsRDDcompute);
            centroids = computeMaster(context, partRes);
        }

        KmeansResult result = new KmeansResult();
        result.centroids = centroids;
        return result;
    }

    private static JavaRDD<InitPartialResult> computeInitLocal(DaalContext context, JavaPairRDD<Integer, CSRNumericTable> dataRDD) {
        return dataRDD.map(new Function<Tuple2<Integer, CSRNumericTable>, InitPartialResult>() {
            public InitPartialResult call(Tuple2<Integer, CSRNumericTable> tup) {

                DaalContext localContext = new DaalContext();

                /* Create an algorithm to initialize the K-Means algorithm on local nodes */
                InitDistributedStep1Local kmeansLocalInit = new InitDistributedStep1Local(localContext, Double.class, InitMethod.randomCSR,
                                                                                          nClusters, nBlocks * nVectorsInBlock,
                                                                                          nVectorsInBlock * tup._1);

                /* Set the input data on local nodes */
                tup._2.unpack(localContext);
                kmeansLocalInit.input.set(InitInputId.data, tup._2);

                /* Initialize the K-Means algorithm on local nodes */
                InitPartialResult pres = kmeansLocalInit.compute();
                pres.pack();
                tup._2.pack();

                localContext.dispose();
                return pres;
            }
        });
    }

    private static InitPartialResult reduceInitPartialResults(DaalContext context, JavaRDD<InitPartialResult> partsRDD) {
        return partsRDD.reduce(new Function2<InitPartialResult, InitPartialResult, InitPartialResult>() {
            public InitPartialResult call(InitPartialResult p1, InitPartialResult p2) {
                DaalContext localContext = new DaalContext();

                /* Create an algorithm to compute new partial result from two partial results */
                InitDistributedStep2Master kmeansMasterInit = new InitDistributedStep2Master(localContext, Double.class,
                                                                                             InitMethod.randomCSR, nClusters);
                /* Set the partial results recieved from the local nodes */
                p1.unpack(localContext);
                p2.unpack(localContext);
                kmeansMasterInit.input.add(InitDistributedStep2MasterInputId.partialResults, p1);
                kmeansMasterInit.input.add(InitDistributedStep2MasterInputId.partialResults, p2);

                /* Compute a new partial result from two partial results */
                InitPartialResult partRes = kmeansMasterInit.compute();
                partRes.pack();

                localContext.dispose();
                return partRes;
            }
        });
    }

    private static HomogenNumericTable computeInitMaster(DaalContext context, InitPartialResult partRes) {
        /* Create an algorithm to compute k-means on the master node */
        InitDistributedStep2Master kmeansMasterInit = new InitDistributedStep2Master(context, Double.class,
                                                                                     InitMethod.randomCSR, nClusters);

        /* Set the reduced partial result to the master algorithm to compute the final result */
        partRes.unpack(context);
        kmeansMasterInit.input.add(InitDistributedStep2MasterInputId.partialResults, partRes);

        /* Compute initial centroids on the master node */
        kmeansMasterInit.compute();

        /* Finalize computations and retrieve the results */
        InitResult initResult = kmeansMasterInit.finalizeCompute();

        return (HomogenNumericTable)initResult.get(InitResultId.centroids);
    }

    private static JavaRDD<PartialResult> computeLocal(DaalContext context, JavaPairRDD<Integer, CSRNumericTable> dataRDD,
            final HomogenNumericTable centroids) {
        return dataRDD.map(new Function<Tuple2<Integer, CSRNumericTable>, PartialResult>() {
            public PartialResult call(Tuple2<Integer, CSRNumericTable> tup) {
                DaalContext localContext = new DaalContext();

                /* Create an algorithm to compute k-means on local nodes */
                DistributedStep1Local kmeansLocal = new DistributedStep1Local(localContext, Double.class,
                                                                              Method.lloydCSR, nClusters);

                /* Set the input data on local nodes */
                tup._2.unpack(localContext);
                centroids.unpack(localContext);
                kmeansLocal.input.set(InputId.data, tup._2);
                kmeansLocal.input.set(InputId.inputCentroids, centroids);

                /* Compute k-means on local nodes */
                PartialResult pres = kmeansLocal.compute();

                pres.pack();
                tup._2.pack();
                centroids.pack();

                localContext.dispose();

                return pres;
            }
        });
    }

    private static PartialResult reducePartialResults(DaalContext context, JavaRDD<PartialResult> partsRDDcompute) {

        /* Reduce partial results computed on local nodes to the single partial result */
        return partsRDDcompute.reduce(new Function2<PartialResult, PartialResult, PartialResult>() {
            public PartialResult call(PartialResult p1, PartialResult p2) {
                DaalContext localContext = new DaalContext();

                /* Create an algorithm to compute k-means on the master node */
                DistributedStep2Master kmeansMaster = new DistributedStep2Master(localContext, Double.class,
                                                                                 Method.lloydCSR, nClusters);
                /* Set the input data on local nodes */
                p1.unpack(localContext);
                p2.unpack(localContext);
                kmeansMaster.input.add(DistributedStep2MasterInputId.partialResults, p1);
                kmeansMaster.input.add(DistributedStep2MasterInputId.partialResults, p2);

                /* Compute k-means on local nodes */
                PartialResult partRes = kmeansMaster.compute();

                partRes.pack();

                localContext.dispose();

                return partRes;
            }
        });
    }

    private static HomogenNumericTable computeMaster(DaalContext context, PartialResult partsRDDcompute) {

        /* Create an algorithm to compute k-means on the master node */
        DistributedStep2Master kmeansMaster = new DistributedStep2Master(context, Double.class,
                                                                         Method.lloydCSR, nClusters);

        /* Set the reduced partial result to the master algorithm to compute the final result */
        partsRDDcompute.unpack(context);
        kmeansMaster.input.add(DistributedStep2MasterInputId.partialResults, partsRDDcompute);

        /* Compute k-means on the master node */
        kmeansMaster.compute();

        /* Finalize computations and retrieve the results */
        Result res = kmeansMaster.finalizeCompute();
        return (HomogenNumericTable)res.get(ResultId.centroids);
    }
}
