# file: covariance_csr_distributed_mpi.py
# ==============================================================
#
# SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
# http://software.intel.com/en-us/articles/intel-sample-source-code-license-agreement/
#
# Copyright 2016 Intel Corporation
#
# THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL PROPERTY RIGHTS.
#
# =============================================================

#
# !  Content:
# !    Pyhton sample of sparse variance-covariance matrix computation in the
# !    distributed processing mode
# !
# !*****************************************************************************

#
## <a name"DAAL-SAMPLE-PY-COVARIANCE_CSR_DISTRIBUTED"></a>
## \example covariance_csr_distributed_mpi.py
#

from os.path import join as jp, realpath, abspath
from os import environ
from mpi4py import MPI

from daal.algorithms.covariance import (
    Distributed_Step1LocalFloat64FastCSR, PartialResult, partialResults,
    Distributed_Step2MasterFloat64FastCSR, data, covariance, mean
)
from daal.data_management import OutputDataArchive, InputDataArchive

import sys
utils_folder = realpath(abspath(jp(environ['DAALROOT'], 'examples', 'python', 'source')))
if utils_folder not in sys.path:
    sys.path.insert(0, utils_folder)
from utils import printNumericTable, createSparseTable

# Input data set parameters
nBlocks = 4
MPI_ROOT = 0

data_dir = jp('data', 'distributed')

datasetFileNames = [
    jp(data_dir, 'covcormoments_csr_1.csv'),
    jp(data_dir, 'covcormoments_csr_2.csv'),
    jp(data_dir, 'covcormoments_csr_3.csv'),
    jp(data_dir, 'covcormoments_csr_4.csv')
]

if __name__ == "__main__":

    comm_size = MPI.COMM_WORLD
    rankId = comm_size.Get_rank()

    # Initialize FileDataSource<CSVFeatureManager> to retrieve the input data from a .csv file
    dataTable = createSparseTable(datasetFileNames[rankId])

    # Create an algorithm to compute a sparse variance-covariance matrix on local nodes
    localAlgorithm = Distributed_Step1LocalFloat64FastCSR()

    # Set the input data set to the algorithm
    localAlgorithm.input.set(data, dataTable)

    # Compute a sparse variance-covariance matrix
    pres = localAlgorithm.compute()

    # Serialize partial results required by step 2
    dataArch = InputDataArchive()
    pres.serialize(dataArch)
    perNodeArchLength = dataArch.getSizeOfArchive()

    nodeResults = dataArch.getArchiveAsArray()

    # Transfer partial results to step 2 on the root node
    data = comm_size.gather(nodeResults, MPI_ROOT)

    if rankId == MPI_ROOT:
        # Create an algorithm to compute a sparse variance-covariance matrix on the master node
        masterAlgorithm = Distributed_Step2MasterFloat64FastCSR()

        for i in range(nBlocks):
            # Deserialize partial results from step 1
            dataArch = OutputDataArchive(data[i])

            dataForStep2FromStep1 = PartialResult()

            dataForStep2FromStep1.deserialize(dataArch)

            # Set local partial results as input for the master-node algorithm
            masterAlgorithm.input.add(partialResults, dataForStep2FromStep1)


        # Merge and finalizeCompute a sparse variance-covariance matrix on the master node
        masterAlgorithm.compute()
        result = masterAlgorithm.finalizeCompute() # Retrieve the algorithm results

        # Print the results
        printNumericTable(result.get(covariance), "Covariance matrix (upper left square 10*10) :", 10, 10)
        printNumericTable(result.get(mean),       "Mean vector:", 1, 10)
