# file: implicit_als_csr_distributed_mpi.py
# ==============================================================
#
# SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
# http://software.intel.com/en-us/articles/intel-sample-source-code-license-agreement/
#
# Copyright 2016 Intel Corporation
#
# THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL PROPERTY RIGHTS.
#
# =============================================================

#
# !  Content:
# !    Python example of the implicit alternating least squares (ALS) algorithm in
# !    the distributed processing mode.
# !
# !    The program trains the implicit ALS model on a training data set.
# !*****************************************************************************

#
## <a name="DAAL-EXAMPLE-PY-IMPLICIT_ALS_CSR_DISTRIBUTED"></a>
## \example implicit_als_csr_distributed_mpi.py
#
import sys
from os.path import join as jp, realpath, abspath
from os import environ

import numpy as np
from mpi4py import MPI

from daal import step1Local, step2Master, step3Local, step4Local

import daal.algorithms.implicit_als as implicit_als
import daal.algorithms.implicit_als.training as training
import daal.algorithms.implicit_als.training.init as init
import daal.algorithms.implicit_als.prediction.ratings as ratings

from daal.data_management import (
    NumericTableIface, KeyValueDataCollection, HomogenNumericTable,
    HomogenNumericTable_Intc, OutputDataArchive, InputDataArchive
)

utils_folder = realpath(abspath(jp(environ['DAALROOT'], 'examples', 'python', 'source')))
if utils_folder not in sys.path:
    sys.path.insert(0, utils_folder)
from utils import createSparseTable, printNumericTable


# Input data set parameters
nBlocks = 4
MPI_ROOT = 0
DATA_DIR = jp('data', 'distributed')

# Number of observations in training data set blocks
trainDatasetFileNames = [
    jp(DATA_DIR, 'implicit_als_csr_1.csv'),
    jp(DATA_DIR, 'implicit_als_csr_2.csv'),
    jp(DATA_DIR, 'implicit_als_csr_3.csv'),
    jp(DATA_DIR, 'implicit_als_csr_4.csv')
]

# Number of observations in transposed training data set blocks
transposedTrainDatasetFileNames = [
    jp(DATA_DIR, 'implicit_als_trans_csr_1.csv'),
    jp(DATA_DIR, 'implicit_als_trans_csr_2.csv'),
    jp(DATA_DIR, 'implicit_als_trans_csr_3.csv'),
    jp(DATA_DIR, 'implicit_als_trans_csr_4.csv')
]

# Algorithm parameters
nUsers = 46  # Full number of users
nFactors = 2  # Number of factors
maxIterations = 5  # Number of iterations in the implicit ALS training algorithm

nodeCPs = [0] * nBlocks
displs = [0] * nBlocks
rdispls = [0] * nBlocks
sdispls = [0] * nBlocks
perNodeCPLength = [0] * nBlocks
perNodeArchLengthsRecv = [0] * nBlocks

dataTable = None
transposedDataTable = None

itemsPartialResultLocal = None
itemsPartialResultsMaster = [0] * nBlocks

step1LocalResult = None
step1LocalResults = [0] * nBlocks
step2MasterResult = None
step3LocalResult = [0] * nBlocks
step4LocalInput = None

all2all_chunks = [0] * nBlocks
gatherItems_chunks = [0] * nBlocks
gatherStep1_chunks = [0] * nBlocks
serializedRecvData = None
serializedData = None

usersPartition = [0] * (nBlocks + 1)
itemsPartition = [0] * (nBlocks + 1)

usersPartialResultLocal = None

predictedRatingsMaster = [[0] * nBlocks for i in range(nBlocks)]


def readData():
    global dataTable, transposedDataTable

    # Read trainDatasetFileName from a file and create a numeric table to store the input data
    dataTable = createSparseTable(rowFileName)
    # Read trainDatasetFileName from a file and create a numeric table to store the input data
    transposedDataTable = createSparseTable(colFileName)


def computeOutBlocks(nRows, dataBlock, dataBlockPartition):

    _, colIndices, rowOffsets = dataBlock.getArrays()

    blockIdFlags = [0] * (nRows * nBlocks)

    for i in range(nRows):
        for j in range(int(rowOffsets[i] - 1), int(rowOffsets[i + 1] - 1)):
            for k in range(1, nBlocks + 1):
                if dataBlockPartition[k - 1] <= colIndices[j] - 1 and colIndices[j] - 1 < dataBlockPartition[k]:
                    blockIdFlags[(k - 1) * nRows + i] = 1

    nNotNull = [0] * nBlocks
    for i in range(nBlocks):
        nNotNull[i] = 0
        for j in range(nRows):
            nNotNull[i] += blockIdFlags[i * nRows + j]

    result = KeyValueDataCollection()

    for i in range(nBlocks):
        indicesTable = HomogenNumericTable_Intc(1, nNotNull[i], NumericTableIface.doAllocate)
        indices = indicesTable.getArray()
        indexId = 0

        for j in range(nRows):
            if blockIdFlags[i * nRows + j]:
                indices[indexId] = j
                indexId += 1

        result[i] = indicesTable

    return result


def initializeModel():
    global itemsPartialResultLocal

    # Create an algorithm object to initialize the implicit ALS model with the default method
    initAlgorithm = init.Distributed(step1Local, method=init.fastCSR)
    initAlgorithm.parameter.fullNUsers = nUsers
    initAlgorithm.parameter.nFactors = nFactors
    initAlgorithm.parameter.seed += rankId
    # Pass a training data set and dependent values to the algorithm
    initAlgorithm.input.set(init.data, transposedDataTable)

    # Initialize the implicit ALS model
    pres = initAlgorithm.compute()
    partialModelLocal = pres.get(init.partialModel)

    itemsPartialResultLocal = training.DistributedPartialResultStep4()
    itemsPartialResultLocal.set(training.outputOfStep4ForStep1, partialModelLocal)


def computeStep1Local(partialResultLocal):
    global step1LocalResult

    # Create algorithm objects to compute a dense variance-covariance matrix in the distributed processing mode using the default method
    algorithm = training.Distributed(step1Local)
    algorithm.parameter.nFactors = nFactors

    # Set input objects for the algorithm
    algorithm.input.set(training.partialModel, partialResultLocal.get(training.outputOfStep4ForStep1))

    # Compute partial estimates on local nodes
    step1LocalResult = algorithm.compute()


def computeStep2Master():
    global step2MasterResult

    # Create algorithm objects to compute a dense variance-covariance matrix in the distributed processing mode using the default method
    algorithm = training.Distributed(step2Master)
    algorithm.parameter.nFactors = nFactors

    # Set input objects for the algorithm
    for i in range(nBlocks):
        algorithm.input.add(training.inputOfStep2FromStep1, step1LocalResults[i])

    # Compute a partial estimate on the master node from the partial estimates on local nodes
    pres = algorithm.compute()

    step2MasterResult = pres.get(training.outputOfStep2ForStep4)


def computeStep3Local(offset, partialResultLocal, outBlocks):
    global step3LocalResult, perNodeCPLength, nodeCPs

    algorithm = training.Distributed(step3Local)
    algorithm.parameter.nFactors = nFactors

    offsetArray = np.array([[offset]], dtype=np.float64)
    offsetTable = HomogenNumericTable(offsetArray)

    algorithm.input.setModel(training.partialModel, partialResultLocal.get(training.outputOfStep4ForStep3))
    algorithm.input.setCollection(training.partialModelBlocksToNode, outBlocks)
    algorithm.input.setTable(training.offset, offsetTable)

    pres = algorithm.compute()

    step3LocalResult = pres.get(training.outputOfStep3ForStep4)

    # MPI_Alltoallv to populate step4LocalInput
    for i in range(nBlocks):
        nodeCPs[i] = serializeDAALObject(step3LocalResult[i])
        perNodeCPLength[i] = len(nodeCPs[i])

    all2all(nodeCPs, perNodeCPLength)


def computeStep4Local(dataTable):

    algorithm = training.Distributed(step4Local)
    algorithm.parameter.nFactors = nFactors

    algorithm.input.setModels(training.partialModels, step4LocalInput)
    algorithm.input.setTable(training.partialData, dataTable)
    algorithm.input.setTable(training.inputOfStep4FromStep2, step2MasterResult)

    pres = algorithm.compute()

    return pres


def trainModel():
    global step2MasterResult, usersPartialResultLocal, itemsPartialResultLocal

    for iteration in range(maxIterations):
        computeStep1Local(itemsPartialResultLocal)

        nodeResults = serializeDAALObject(step1LocalResult)

        # Gathering step1LocalResult on the master
        gatherStep1(nodeResults)
        crossProductBuf = None

        if rankId == MPI_ROOT:
            computeStep2Master()
            crossProductBuf = serializeDAALObject(step2MasterResult)

        crossProductBuf = comm.bcast(crossProductBuf)

        step2MasterResult = deserializeNumericTable(crossProductBuf)

        computeStep3Local(itemsPartition[rankId], itemsPartialResultLocal, itemsOutBlocks)

        usersPartialResultLocal = computeStep4Local(dataTable)

        computeStep1Local(usersPartialResultLocal)

        nodeResults = serializeDAALObject(step1LocalResult)
        # Gathering step1LocalResult on the master
        gatherStep1(nodeResults)

        if rankId == MPI_ROOT:
            computeStep2Master()
            crossProductBuf = serializeDAALObject(step2MasterResult)

        crossProductBuf = comm.bcast(crossProductBuf)
        step2MasterResult = deserializeDAALObject(crossProductBuf)

        computeStep3Local(usersPartition[rankId], usersPartialResultLocal, usersOutBlocks)
        itemsPartialResultLocal = computeStep4Local(transposedDataTable)

    # Gather all itemsPartialResultLocal to itemsPartialResultsMaster on the master and distributing the result over other ranks
    nodeResults = serializeDAALObject(itemsPartialResultLocal)
    gatherItems(nodeResults)


def testModel():
    predictedRatingsLocal = [0] * nBlocks

    # Create an algorithm object to predict recommendations of the implicit ALS model
    for i in range(nBlocks):
        # perNodeArchLength = None
        algorithm = ratings.Distributed_Step1LocalFloat32AllUsersAllItems()
        algorithm.parameter.nFactors = nFactors

        algorithm.input.set(ratings.usersPartialModel, usersPartialResultLocal.get(training.outputOfStep4ForStep1))
        algorithm.input.set(ratings.itemsPartialModel, itemsPartialResultsMaster[i].get(training.outputOfStep4ForStep1))

        res = algorithm.compute()

        predictedRatingsLocal[i] = res.get(ratings.prediction)

        nodeResults = serializeDAALObject(predictedRatingsLocal[i])
        gatherPrediction(nodeResults, i)


def gatherStep1(nodeResults):
    global step1LocalResults, displs, gatherStep1_chunks, serializedData

    perNodeArchLengthMaster = comm.gather(len(nodeResults))

    if rankId == MPI_ROOT:
        memoryBuf = 0
        for i in range(nBlocks):
            memoryBuf += perNodeArchLengthMaster[i]
        serializedData = np.zeros(memoryBuf, dtype=np.uint8)

        shift = 0
        for i in range(nBlocks):
            displs[i] = shift
            shift += perNodeArchLengthMaster[i]

    # Transfer partial results to step 2 on the root node
    # serializedData = comm.gather(nodeResults)
    comm.Gatherv([nodeResults, len(nodeResults), MPI.CHAR],
                 [serializedData, perNodeArchLengthMaster, displs, MPI.CHAR])

    if rankId == MPI_ROOT:
        for i in range(nBlocks):
            # Deserialize partial results from step 1
            start = displs[i]
            end = sum(perNodeArchLengthMaster[:i + 1])
            gatherStep1_chunks[i] = serializedData[start:end]

            dataArch = OutputDataArchive(gatherStep1_chunks[i])
            dataTable = training.DistributedPartialResultStep1()
            dataTable.deserialize(dataArch)
            step1LocalResults[i] = dataTable


def gatherItems(nodeResults):
    global itemsPartialResultsMaster, displs, gatherItems_chunks

    perNodeArchLengthMaster = comm.allgather(len(nodeResults))

    memoryBuf = 0
    for i in range(nBlocks):
        memoryBuf += perNodeArchLengthMaster[i]
    serializedData = np.zeros(memoryBuf, dtype=np.uint8)

    shift = 0
    for i in range(nBlocks):
        displs[i] = shift
        shift += perNodeArchLengthMaster[i]

    # Transfer partial results to step 2 on the root node
    comm.Allgatherv([nodeResults, len(nodeResults), MPI.CHAR],
                    [serializedData, perNodeArchLengthMaster, displs, MPI.CHAR])

    for i in range(nBlocks):
        # Deserialize partial results from step 4
        start = displs[i]
        end = sum(perNodeArchLengthMaster[:i + 1])
        gatherItems_chunks[i] = serializedData[start:end]
        itemsPartialResultsMaster[i] = deserializeStep4PartialResult(gatherItems_chunks[i])


def gatherPrediction(nodeResults, iNode):
    global predictedRatingsMaster

    # Transfer partial results to step 2 on the root node
    serializedData = comm.gather(nodeResults)

    if rankId == MPI_ROOT:
        for i in range(nBlocks):
            # Deserialize partial results from step 1
            outArch = OutputDataArchive(serializedData[i])
            res = ratings.Result()
            res.deserialize(outArch)
            predictedRatingsMaster[iNode][i] = res


def all2all(nodeResults, perNodeArchLengths):
    global sdispls, perNodeArchLengthsRecv, step4LocalInput, serializedRecvData, all2all_chunks

    memoryBuf = 0
    shift = 0
    for i in range(nBlocks):
        memoryBuf += perNodeArchLengths[i]
        sdispls[i] = shift
        shift += perNodeArchLengths[i]

    # memcpy to avoid double compute
    serializedSendData = np.array([], dtype=np.uint8)
    memoryBuf = 0
    for i in range(nBlocks):
        serializedSendData = np.concatenate((serializedSendData, nodeResults[i]))
        nodeResults[i] = None

    perNodeArchLengthsRecv = comm.alltoall(perNodeArchLengths)

    memoryBuf = 0
    shift = 0
    for i in range(nBlocks):
        memoryBuf += perNodeArchLengthsRecv[i]
        rdispls[i] = shift
        shift += perNodeArchLengthsRecv[i]

    serializedRecvData = np.zeros(memoryBuf, dtype=np.uint8)
    # Transfer partial results to step 2 on the root node
    comm.Alltoallv([serializedSendData, perNodeArchLengths, sdispls, MPI.CHAR],
                   [serializedRecvData, perNodeArchLengthsRecv, rdispls, MPI.CHAR])

    for i in range(nBlocks):
        start = rdispls[i]
        end = sum(perNodeArchLengthsRecv[:i + 1])
        all2all_chunks[i] = serializedRecvData[start:end]
        step4LocalInput[i] = deserializePartialModel(all2all_chunks[i])


def serializeDAALObject(data):
    # Create a data archive to serialize the numeric table
    dataArch = InputDataArchive()

    # Serialize the numeric table into the data archive
    data.serialize(dataArch)

    length = dataArch.getSizeOfArchive()
    buffer = np.zeros(length, dtype=np.uint8)
    dataArch.copyArchiveToArray(buffer)
    return buffer


def deserializeDAALObject(buffer):
    # Create a data archive to deserialize the numeric table
    dataArch = OutputDataArchive(buffer)

    # Create a numeric table object
    dataTable = HomogenNumericTable()

    # Deserialize the numeric table from the data archive
    dataTable.deserialize(dataArch)

    return dataTable

def deserializeNumericTable(buffer):
    outDataArch = OutputDataArchive(buffer)

    restoredDataTable = HomogenNumericTable()

    restoredDataTable.deserialize(outDataArch)

    return restoredDataTable

def deserializePartialModel(buffer):

    dataArch = OutputDataArchive(buffer)

    data = implicit_als.PartialModel()

    data.deserialize(dataArch)

    return data


def deserializeStep4PartialResult(buffer):
    outArch = OutputDataArchive(buffer)

    pres = training.DistributedPartialResultStep4()

    pres.deserialize(outArch)

    return pres


if __name__ == "__main__":

    comm = MPI.COMM_WORLD
    rankId = comm.Get_rank()

    rowFileName = trainDatasetFileNames[rankId]
    colFileName = transposedTrainDatasetFileNames[rankId]

    readData()
    initializeModel()

    step4LocalInput = KeyValueDataCollection()
    itemsPartialResultPrediction = KeyValueDataCollection()

    userNum = dataTable.getNumberOfRows()
    itemNum = transposedDataTable.getNumberOfRows()

    usersPartitionTmp = comm.allgather(userNum)
    itemsPartitionTmp = comm.allgather(itemNum)

    usersPartition[0] = 0
    itemsPartition[0] = 0
    for i in range(nBlocks):
        usersPartition[i + 1] = usersPartition[i] + usersPartitionTmp[i]
        itemsPartition[i + 1] = itemsPartition[i] + itemsPartitionTmp[i]

    usersOutBlocks = computeOutBlocks(dataTable.getNumberOfRows(), dataTable, itemsPartition)
    itemsOutBlocks = computeOutBlocks(transposedDataTable.getNumberOfRows(), transposedDataTable, usersPartition)

    trainModel()
    testModel()

    print("Compute OK")

    # if rankId == MPI_ROOT:
    #     for i in range(nBlocks):
    #         for j in range(nBlocks):
    #             print("prediction {}, {}".format(i, j))
    #             printNumericTable(predictedRatingsMaster[i][j].get(ratings.prediction))
