# file: multinomial_naive_bayes_csr_distributed_mpi.py
# ==============================================================
#
# SAMPLE SOURCE CODE - SUBJECT TO THE TERMS OF SAMPLE CODE LICENSE AGREEMENT,
# http://software.intel.com/en-us/articles/intel-sample-source-code-license-agreement/
#
# Copyright 2016 Intel Corporation
#
# THIS FILE IS PROVIDED "AS IS" WITH NO WARRANTIES, EXPRESS OR IMPLIED, INCLUDING BUT
# NOT LIMITED TO ANY IMPLIED WARRANTY OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
# PURPOSE, NON-INFRINGEMENT OF INTELLECTUAL PROPERTY RIGHTS.
#
# =============================================================

#
# !  Content:
# !    Python sample of Naive Bayes classification in the distributed processing
# !    mode.
# !
# !    The program trains the Naive Bayes model on a supplied training data set
# !    and then performs classification of previously unseen data.
# !*****************************************************************************

#
## <a name="DAAL-SAMPLE-PY-MULTINOMIAL_NAIVE_BAYES_CSR_DISTRIBUTED"></a>
## \example multinomial_naive_bayes_csr_distributed_mpi.py
#

import sys
from os import environ
from os.path import join as jp, realpath, abspath

from mpi4py import MPI

import daal.algorithms.classifier as classifier
import daal.algorithms.multinomial_naive_bayes.prediction as prediction
import daal.algorithms.multinomial_naive_bayes.training as training

from daal.data_management import (
    DataSourceIface, FileDataSource, OutputDataArchive, InputDataArchive
)
utils_folder = realpath(abspath(jp(environ['DAALROOT'], 'examples', 'python', 'source')))
if utils_folder not in sys.path:
    sys.path.insert(0, utils_folder)
from utils import printNumericTables, createSparseTable

DATA_PREFIX = jp('data', 'distributed')

# Input data set parameters
trainDatasetFileNames = [
    jp(DATA_PREFIX, 'naivebayes_train_csr.csv'),
    jp(DATA_PREFIX, 'naivebayes_train_csr.csv'),
    jp(DATA_PREFIX, 'naivebayes_train_csr.csv'),
    jp(DATA_PREFIX, 'naivebayes_train_csr.csv')
]
trainGroundTruthFileNames = [
    jp(DATA_PREFIX, 'naivebayes_train_labels.csv'),
    jp(DATA_PREFIX, 'naivebayes_train_labels.csv'),
    jp(DATA_PREFIX, 'naivebayes_train_labels.csv'),
    jp(DATA_PREFIX, 'naivebayes_train_labels.csv')
]

testDatasetFileName = jp(DATA_PREFIX, 'naivebayes_test_csr.csv')
testGroundTruthFileName = jp(DATA_PREFIX, 'naivebayes_test_labels.csv')

nClasses = 20
nBlocks = 4

MPI_ROOT = 0

trainingResult = None
predictionResult = None


def trainModel():
    global trainingResult

    # Retrieve the input data from a .csv file
    trainDataTable = createSparseTable(trainDatasetFileNames[rankId])

    # Initialize FileDataSource to retrieve the input data from a .csv file
    trainLabelsSource = FileDataSource(trainGroundTruthFileNames[rankId],
                                       DataSourceIface.doAllocateNumericTable,
                                       DataSourceIface.doDictionaryFromContext)

    # Retrieve the data from input files
    trainLabelsSource.loadDataBlock()

    # Create an algorithm object to train the Naive Bayes model based on the local-node data
    localAlgorithm = training.Distributed_Step1LocalFloat64FastCSR(nClasses)

    # Pass a training data set and dependent values to the algorithm
    localAlgorithm.input.set(classifier.training.data, trainDataTable)
    localAlgorithm.input.set(classifier.training.labels, trainLabelsSource.getNumericTable())

    # Train the Naive Bayes model on local nodes
    pres = localAlgorithm.compute()

    # Serialize partial results required by step 2
    dataArch = InputDataArchive()
    pres.serialize(dataArch)

    nodeResults = dataArch.getArchiveAsArray()

    # Transfer partial results to step 2 on the root node
    serializedData = comm.gather(nodeResults)

    if rankId == MPI_ROOT:
        # Create an algorithm object to build the final Naive Bayes model on the master node
        masterAlgorithm = training.Distributed_Step2MasterFloat64FastCSR(nClasses)

        for i in range(nBlocks):
            # Deserialize partial results from step 1
            dataArch = OutputDataArchive(serializedData[i])

            dataForStep2FromStep1 = classifier.training.PartialResult()
            dataForStep2FromStep1.deserialize(dataArch)

            # Set the local Naive Bayes model as input for the master-node algorithm
            masterAlgorithm.input.add(classifier.training.partialModels, dataForStep2FromStep1)

        # Merge and finalizeCompute the Naive Bayes model on the master node
        masterAlgorithm.compute()
        trainingResult = masterAlgorithm.finalizeCompute()


def testModel():
    global predictionResult

    # Retrieve the input data from a .csv file
    testDataTable = createSparseTable(testDatasetFileName)

    # Create an algorithm object to predict values of the Naive Bayes model
    algorithm = prediction.Batch_Float64FastCSR(nClasses)

    # Pass a testing data set and the trained model to the algorithm
    algorithm.input.setTable(classifier.prediction.data, testDataTable)
    algorithm.input.setModel(classifier.prediction.model, trainingResult.get(classifier.training.model))

    # Predict values of the Naive Bayes model
    # Result class from classifier.prediction
    predictionResult = algorithm.compute()


def printResults():

    testGroundTruth = FileDataSource(testGroundTruthFileName,
                                     DataSourceIface.doAllocateNumericTable,
                                     DataSourceIface.doDictionaryFromContext)
    testGroundTruth.loadDataBlock()

    printNumericTables(testGroundTruth.getNumericTable(),
                       predictionResult.get(classifier.prediction.prediction),
                       "Ground truth",
                       "Classification results",
                       "NaiveBayes classification results (first 20 observations):",
                       20,
                       interval=15,
                       flt64=False)

if __name__ == "__main__":

    comm = MPI.COMM_WORLD
    rankId = comm.Get_rank()

    trainModel()

    if rankId == MPI_ROOT:
        testModel()
        printResults()
